// Tony Urschler
// November 2, 2021

#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include "config.h"
#include "oss.h"

/* VERBOSE */
int verbose = 1;

// prototypes
void killandfree();
void handle_sigint(int sig);
static void handle_timer(int sig);
static int setupinterrupt(void);
static int setuptimer(int time);
int req_lt_avail(const int* req, const int* avail,
        const int pnum, const int num_res);
int deadlock(const int* available, const int m, const int n,
        const int* request, const int* allocated);
void insertpid(int lpid, int index, int resource);
struct node* removepid(int lpid);
void logmsg(char* msg, int opt);
void printresourcetable(void);

// globals
int segment_id[3];
int* sysclock;
rd_t* rds;
rdoverview_t* rdo;
int semid;
long A[1], B[1];
pid_t kiddos[18];
int msgid;
struct node *head = NULL;
struct node *current = NULL;
int numoflines = 0;
int termbydeadlock = 0;
int termnormally = 0;
int immediategrants = 0;
int waitgrants = 0;
int deadlockdetectionruns = 0;

int main(int argc, char *argv[]) {
    signal(SIGINT, handle_sigint);
    signal(SIGCHLD, SIG_IGN);

    // setting error message
    char errmsgraw[100];
    snprintf(errmsgraw, sizeof(errmsgraw), "%s: Error", argv[0]);
    char* errmsg = errmsgraw+2;

    // allocating shared memory for system clock
    segment_id[0] = shmget(IPC_PRIVATE, 2*sizeof(int),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    sysclock = (int*) shmat(segment_id[0], NULL, 0);

    // allocating shared memory for resource descriptors
    segment_id[1] = shmget(IPC_PRIVATE, 20*sizeof(rd_t),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    rds = (rd_t*) shmat(segment_id[1], NULL, 0);

    // allocating shared memory for resource overview
    segment_id[2] = shmget(IPC_PRIVATE, sizeof(rdoverview_t),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    rdo = (rdoverview_t*) shmat(segment_id[2], NULL, 0);

    // initializing msg queue
    key_t key;
    key = ftok("testprog", 65);
    msgid = msgget(key, 0666 | IPC_CREAT);

    // sysclock[0] holds seconds and sysclock[1] holds nanoseconds
    sysclock[0] = 0;
    sysclock[1] = 0;

    // initializing all resource descriptors
    srand(getpid());
    int numofsharable = (rand() % (5 - 3 + 1)) + 3;
    for (int i = 0; i < 20; i++) {
        rds[i].isSharable = (i < numofsharable) ? 1 : 0;
        rds[i].initnum = (rand() % (10 - 1 + 1)) + 1;
        rds[i].numremaining = rds[i].initnum;
        rdo->available[i] = rds[i].numremaining;
    }
    for (int i = 0; i < 360; i++) {
        rdo->allocation[i] = 0;
        rdo->need[i] = 0;
        rdo->request[i] = 0;
    }

    // setting up semaphore
    semid = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600);
    if (semid == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": License failed to initialize");
        char* initerr = errmsgraw;
        perror(initerr);
        killandfree();
    }
    // set semaphore value
    int rc;
    union semun sem_val;
    sem_val.val = 1;
    rc = semctl(semid, 0, SETVAL, sem_val);
    if (rc ==-1) {
        perror("setting sem val:");
        killandfree();
    }
    // creating var for operations
    struct sembuf sem_op;
    sem_op.sem_num = 0;
    sem_op.sem_flg = 0;

    // setting up timer
    if (setupinterrupt() == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup handler for SIGPROF");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }
    if (setuptimer(5) == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup ITIMER_PROF interval timer");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }

    // initializing bit array
    A[0] = 0;

    // making ids into strings for children
    char resDesBuff[50], overviewBuff[50], clockbuff[50], msgIdBuff[10], semIdBuff[10];
    snprintf(resDesBuff, 49, "%d", segment_id[1]);
    snprintf(overviewBuff, 49, "%d", segment_id[2]);
    snprintf(clockbuff, 49, "%d", segment_id[0]);
    snprintf(semIdBuff, 9, "%d", semid);
    char* rds_id_str = resDesBuff;
    char* rdo_id_str = overviewBuff;
    char* clock_id_str = clockbuff;
    char* sem_id_str = semIdBuff;
    int makebabyns = 0;
    int makebabysec = 0;
    int numberofproc = 0;
    int totalproc = 0;
    int sizeofwaitlist = 0;
    int totalgrants = 0;
    int rounds = 0;
    char logbuffer[200];
    char* logmsg_str;

    // for sending/receiving messages
    mymsg_t mymsg;

    logmsg("OSS Starting...", 1);
    numoflines++;

    // starting main loop
    for (;;) {
        //check termination criteria
        if (numberofproc == 0 && totalproc == 40)
            break;
        if (numoflines >= 100000)
            break;

        // go through the wait list
        if (head != NULL) {
            struct node *tmp = head;
            int count = 0;
            do {
                // check if there's enough available for the proc's request
                if (rdo->request[tmp->requestindex] <= rdo->available[tmp->resourcenumber]) {
                    rdo->available[tmp->resourcenumber] -= rdo->request[tmp->requestindex];
                    rdo->allocation[tmp->requestindex] += rdo->request[tmp->requestindex];
                    rdo->need[tmp->requestindex] -= rdo->request[tmp->requestindex];
                    
                    deadlockdetectionruns++;
                    // run deadlock algorithm and see if can allocate now
                    if (deadlock(rdo->available, 20, 18, rdo->need, rdo->allocation)) {
                        rdo->available[tmp->resourcenumber] += rdo->request[tmp->requestindex];
                        rdo->allocation[tmp->requestindex] -= rdo->request[tmp->requestindex];
                        rdo->need[tmp->requestindex] += rdo->request[tmp->requestindex];
                        count++;
                    }
                    
                    // if its safe for the process to take the resource
                    else {
                            rdo->request[tmp->requestindex] = 0;
                            mymsg.mtype = tmp->lpid + 2;
                            strcpy(mymsg.mtext, "GO");
                            msgsnd(msgid, &mymsg, sizeof(mymsg), 0);
                            struct node *dlt = removepid(tmp->lpid);
                            free(dlt);
                            int randinc = rand() % 10000;
                            sysclock[1] += randinc;
                            sizeofwaitlist--;
                            waitgrants++;
                    }
                }
                // go to next node
                tmp = tmp->next;
                // if there is a deadlock in the system
                if (count == numberofproc) {
                    // go through all the processes and find the one with the most resources
                    int maxindex = 0;
                    int max = 0;
                    for (int k = 0; k < 18; k++) {
                        int tally = 0;
                        for (int j = 0; j < 20; j++) {
                            tally += rdo->allocation[20*k+j];
                        }
                        if (tally > max) {
                            max = tally;
                            maxindex = k;
                        }
                    }
                    tmp = head;
                    termbydeadlock++;
                    // kill child and clear its memory
                }
//                printf("sizeofwaitlist=%d numberofproc=%d\n", sizeofwaitlist, numberofproc);
            } while (tmp != NULL);
        }
        int i;

        // wait for clock access
        sem_op.sem_op = -1;
        semop(semid, &sem_op, 1);
        
        // check if its time to make another child
        if ((numberofproc == 0 || sysclock[0] > makebabysec
            || (sysclock[0] == makebabysec && sysclock [1] >= makebabyns))
            && totalproc < 40) {
            
            // find an empty bit
            for (i = 0; i < 18; i++)
                if (TestBit(A, i) == 0)
                    break;

            // make child if there's room
            if (i < 18) {
                SetBit(A, i);
                char lpidBuff[5];
                snprintf(lpidBuff, 4, "%d", i);
                char* lpid_str = lpidBuff;
                numberofproc++;
                totalproc++;
                snprintf(msgIdBuff, 9, "%d", msgid);
                char* msg_id_str = msgIdBuff;
                kiddos[i] = fork();
                if (kiddos[i] == 0)
                    execlp("./user_proc", "./user_proc", rds_id_str,
                            clock_id_str, msg_id_str, lpid_str, 
                            rdo_id_str, sem_id_str, NULL);
            }
            // log child was created
            if (verbose) {
                char bufferkid[100];
                snprintf(bufferkid, 99, "OSS: Created child with PID %d at time %d:%d",
                        i, sysclock[0], sysclock[1]);
                char* message = bufferkid;
                logmsg(message, 0);
                numoflines++;
            }

            // determine new time to make process makebabysec
            makebabysec = sysclock[0];
            makebabyns = (rand() % 500000000) + sysclock[1];
            if (makebabyns >= 1000000000) {
                makebabysec += 1;
                makebabyns -= 1000000000;
            }
        }
        // give up access to clock
        sem_op.sem_op = 1;
        semop(semid, &sem_op, 1);
        char buffer[100];

        // parent waiting for a request
        int x = msgrcv(msgid, &mymsg, sizeof(mymsg), 22, IPC_NOWAIT);
        
        if (x == -1) {
            // increment system clock [0,1] sec and [0,1] millisec
            sem_op.sem_op = -1;
            semop(semid, &sem_op, 1);
            sysclock[0] += rand() % 2;
            sysclock[1] += rand() % 1000000;
            sem_op.sem_op = 1;
            semop(semid, &sem_op, 1);
            rounds++;
            if (rounds >= 20)
                break;
            continue;
        }
        rounds = 0;

        // parse message from child
        int lpid = atoi(strtok(mymsg.mtext, " "));
        char *requesttype = strtok(NULL, " ");
        int resource = atoi(strtok(NULL, " "));

        // if it is a resource request
        if (requesttype[0] == 'R') {
            int index = lpid * 20 + resource;
            sem_op.sem_op = -1;
            semop(semid, &sem_op, 1);
            snprintf(logbuffer, 199, "OSS: Child %d requested R%d at time %d:%d",
                    lpid, resource, sysclock[0], sysclock[1]);
            char* logmsg_str = logbuffer;
            logmsg(logmsg_str, 0);
            numoflines++;
            sem_op.sem_op = 1;
            semop(semid, &sem_op, 1);
            
            // if the resource is sharable
            if (rds[resource].isSharable == 1) {
                mymsg.mtype = lpid + 2;
                strcpy(mymsg.mtext, "GO");
                msgsnd(msgid, &mymsg, sizeof(mymsg), 0);
                rdo->allocation[index] += rdo->request[index];
                rdo->request[index] = 0;
                immediategrants++;
                sem_op.sem_op = -1;
                semop(semid, &sem_op, 1);
                snprintf(buffer, 99, "OSS: Request is safe. Allocating %d of resource #%d to Child %d at time %d:%d",
                        rdo->request[index], resource, lpid, sysclock[0], sysclock[1]);
                sem_op.sem_op = 1;
                semop(semid, &sem_op, 1);
                char* message = buffer;
                logmsg(message, 0);
                numoflines++;
                totalgrants++;
                if (verbose && totalgrants % 20 == 0)
                    printresourcetable();
            }

            // not sharable
            else {
                // if request is greater than need
                if (rdo->request[index] > rdo->need[index]) {
                    snprintf(errmsgraw, sizeof(errmsgraw),
                            errmsg, ": Process requested more than they needed");
                    char* requesterr = errmsgraw;
                    perror(requesterr);
                    killandfree();
                }
                char* message;
                // if request is less than available
                if (rdo->request[index] <= rdo->available[resource]) {
                    rdo->available[resource] -= rdo->request[index];
                    rdo->allocation[index] += rdo->request[index];
                    rdo->need[index] -= rdo->request[index];
                    
                    if (verbose) {
                        sem_op.sem_op = -1;
                        semop(semid, &sem_op, 1);
                        snprintf(buffer, 99, "OSS: Running deadlock detection at time %d:%d",
                                sysclock[0], sysclock[1]);
                        sem_op.sem_op = 1;
                        semop(semid, &sem_op, 1);
                        message = buffer;
                        logmsg(message, 0);
                        numoflines++;
                    }

                    deadlockdetectionruns++;
                    // if unsafe
                    if (deadlock(rdo->available, 20, 18, rdo->need, rdo->allocation)) {
                        //printf("Its not safe to take\n");
                        rdo->available[resource] += rdo->request[index];
                        rdo->allocation[index] -= rdo->request[index];
                        rdo->need[index] += rdo->request[index];
                        // add to wait queue
                        insertpid(lpid, index, resource);
                        sizeofwaitlist++;
    
                        sem_op.sem_op = -1;
                        semop(semid, &sem_op, 1);
                        snprintf(buffer, 99, "OSS: Request lead to an unsafe state. Adding Child %d to waitlist at time %d:%d",
                                lpid, sysclock[0], sysclock[1]);
                        sem_op.sem_op = 1;
                        semop(semid, &sem_op, 1);
                        
                        message = buffer;
                        logmsg(message, 0);
                        numoflines++;
                    }

                    //if safe
                    else {
                        sem_op.sem_op = -1;
                        semop(semid, &sem_op, 1);
                        snprintf(buffer, 99, "OSS: Allocating %d of resource #%d to Child %d at time %d:%d",
                                rdo->request[index], resource, lpid, sysclock[0], sysclock[1]);
                        sem_op.sem_op = 1;
                        semop(semid, &sem_op, 1);

                        message = buffer;
                        logmsg(message, 0);
                        numoflines++;
                        //printf("It's safe to take\n");
                        rdo->request[index] = 0;
                        mymsg.mtype = lpid + 2;
                        strcpy(mymsg.mtext, "GO");
                        msgsnd(msgid, &mymsg, sizeof(mymsg), 0);
                        immediategrants++;
                        totalgrants++;
                        if (verbose && totalgrants % 20 == 0)
                            printresourcetable();
                    }
                }
                // if request is less than available
                else {
                    // add to wait queue
                    insertpid(lpid, index, resource);
                    sizeofwaitlist++;
                    
                    sem_op.sem_op = -1;
                    semop(semid, &sem_op, 1);
                    snprintf(buffer, 99, "OSS: Not enough resources to grand request. Adding Child %d to waitlist at time %d:%d", lpid, sysclock[0], sysclock[1]);
                    sem_op.sem_op = 1;
                    semop(semid, &sem_op, 1);

                    message = buffer;
                    logmsg(message, 0);
                    numoflines++;
                }
            }
        }

        // if the process gave up a resource
        else if (requesttype[0] == 'G') {
            sem_op.sem_op = -1;
            semop(semid, &sem_op, 1);
            snprintf(logbuffer, 199, "OSS: Child %d giving up R%d at time %d:%d",
                    lpid, resource, sysclock[0], sysclock[1]);
            char* logmsg_str = logbuffer;
            logmsg(logmsg_str, 0);
            numoflines++;
            sem_op.sem_op = 1;
            semop(semid, &sem_op, 1);
            int index = lpid * 20 + resource;
            rdo->need[index] += rdo->allocation[index];
            rdo->available[resource] += rdo->allocation[index];
            rdo->allocation[index] = 0;
            rdo->request[index] = 0;
            mymsg.mtype = lpid + 2;
            strcpy(mymsg.mtext, "GO");
            msgsnd(msgid, &mymsg, sizeof(mymsg), 0);
        }

        // if it is a termination notice
        else if (requesttype[0] == 'T') {
            if (verbose) {
                sem_op.sem_op = -1;
                semop(semid, &sem_op, 1);
                snprintf(logbuffer, 199, "OSS: Child %d terminated at time %d:%d. Deallocating all resources.",
                        lpid, sysclock[0], sysclock[1]);
                char* logmsg_str = logbuffer;
                logmsg(logmsg_str, 0);
                numoflines++;
                sem_op.sem_op = 1;
                semop(semid, &sem_op, 1);
            }

            // deallocate all their resources
            int index = lpid * 20;
            for (int k = 0; k < 20; k++) {
                if (rdo->allocation[index+k] != 0) {
                    rdo->available[k] += rdo->allocation[index+k];
                    rdo->allocation[index+k] = 0;
                    rdo->need[index+k] = 0;
                    rdo->request[index+k] = 0;
                }
            }
            ClearBit(A, lpid);
            numberofproc--;
            termnormally++;
        }

        // increment system clock [0,1] sec and [0,1] millisec
        sem_op.sem_op = -1;
        semop(semid, &sem_op, 1);
        sysclock[0] += rand() % 2;
        sysclock[1] += rand() % 1000000;
        sem_op.sem_op = 1;
        semop(semid, &sem_op, 1);
    }
    killandfree();
}

// frees all shared memory and kill child processes
void killandfree() {
    // killing children
    for (int i = 0; i < 18; i++)
        if (TestBit(A, i) > 0)
            kill(kiddos[i], SIGINT);
    
    // logging stats to file
    char buffer[100];
    snprintf(buffer, 99, "OSS: Finished at time %d:%d", sysclock[0], sysclock[1]);
    char* message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "##########  SUMMARY  ##########");
    message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "Total Number of Lines (logfile): %d", numoflines);
    message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "Number of immediate grants (no waiting): %d",
            immediategrants);
    message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "Number of delayed grants (had to wait): %d",
            waitgrants);
    message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "Number of processes terminated due to deadlock: %d",
            termbydeadlock);
    message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "Number of processes terminated normally: %d",
            termnormally);
    message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "Number of times deadlock detection algorithm was run: %d",
            deadlockdetectionruns);
    message = buffer;
    logmsg(message, 0);

    snprintf(buffer, 99, "Percentage of processes terminated by deadlock: %f",
            (float)termbydeadlock / (termnormally + termbydeadlock));
    message = buffer;
    logmsg(message, 0);

    // deallocating shared memory
    int rc = semctl(semid, 1, IPC_RMID);
    
    shmdt(sysclock);
    shmdt(rds);
    shmctl(segment_id[0], IPC_RMID, 0);
    shmctl(segment_id[1], IPC_RMID, 0);
    shmctl(segment_id[2], IPC_RMID, 0);

    msgctl(msgid, IPC_RMID, NULL);

    exit(EXIT_SUCCESS);
}

// called when SIGINT is received
void handle_sigint(int sig) {
    printf("\nSIGINT caught!\n");
    killandfree();
}

// determines if there's enough resources for a 
// processes request
int req_lt_avail(const int* req, const int* avail,
                  const int pnum, const int num_res) {
    int i = 0;
    
    for (; i < num_res; i++)
        if (req[pnum*num_res+i] > avail[i])
            break;
    return (i == num_res);
}

// determines if there is a possible deadlock
int deadlock(const int* available, const int m, const int n,
              const int* request, const int* allocated) {
    int work[m];    // m resources
    int finish[n];  // n processes

    for (int i = 0; i < m; i++)
        work[i] = available[i];
    for (int i = 0; i< n; i++)
        finish[i] = 0;

    int p = 0;
    for (; p < n; p++) {    // for each process
        if (finish[p]) continue;
        if (req_lt_avail (request, work, p, m)) {
            finish[p] = 1;
            for (int i = 0; i < m; i++)
                work[i] += allocated[p*m+i];
            p = -1;
        }
    }

    for (p = 0; p < n; p++)
        if (!finish[p])
            break;

    return (p != n);
}

//insert node
void insertpid(int lpid, int index, int resource) {
   //create a link
   struct node *link = (struct node*) malloc(sizeof(struct node));
   link->lpid = lpid;
   link->requestindex = index;
   link->resourcenumber = resource;
   link->next = head;
   head = link;
}

//delete node with given pid
struct node* removepid(int lpid) {
   struct node* current = head;
   struct node* previous = NULL;
   if(head == NULL) {
      return NULL;
   }
   while(current->lpid != lpid) {
      if(current->next == NULL) {
         return NULL;
      } else {
         previous = current;
         current = current->next;
      }
   }
   if(current == head) {
      head = head->next;
   } else {
      previous->next = current->next;
   }    
   return current;
}

// used to log messages to a specified file
void logmsg(char* msg, int opt) {
    if (opt != 1) {
        FILE *fptr;
        fptr = fopen("logfile", "a");
        fprintf(fptr, "%s\n", msg);
        fclose(fptr);
    }
    else {
        FILE *fptr;
        fptr = fopen("logfile", "w");
        fprintf(fptr, "%s\n", msg);
        fclose(fptr);
    }
}

// called when the timer interval has occurred
void handle_timer(int sig) {
    printf("Times up! Killing children and freeing shared memory.\n");
    killandfree();
}

// sets up the interrupt for timer interval
static int setupinterrupt(void) {
    struct sigaction act;
    act.sa_handler = handle_timer;
    act.sa_flags = 0;
    return (sigemptyset(&act.sa_mask) || sigaction(SIGALRM, &act, NULL));
}

// sets up timer
static int setuptimer(int time) {
    struct itimerval value;
    value.it_interval.tv_sec = time;
    value.it_interval.tv_usec = 0;
    value.it_value = value.it_interval;
    return (setitimer(ITIMER_REAL, &value, NULL));
}

// prints the resources table
void printresourcetable(void) {
    FILE *fptr;
    fptr = fopen("logfile", "a");
    fprintf(fptr, "   ");
    for (int i = 0; i < 20; i++)
        fprintf(fptr, " R%2d", i);
    fprintf(fptr, "\n");
    for (int i = 0; i < 18; i++) {
        fprintf(fptr, "P%2d", i);
        for (int j = 0; j < 20; j++)
            fprintf(fptr, " %3d", rdo->allocation[i*20+j]);
        fprintf(fptr, "\n");
    }
    numoflines += 19;
    fclose(fptr);
}
